
//  =================================================================
//  EXAMPLE.C                   2012-02-19 adapted by TG
//
//  Example code fragments for:
//   - accessing TerraScan published functions
//  =================================================================

#include <mssystem.fdf>
#include <mscexpr.fdf>
#include <msoutput.fdf>
#include <mswindow.fdf>
#include <toolsubs.h>
#include <w32dde.h>
//#include <msmbstr.fdf>

typedef struct {
    BYTE    red ;
    BYTE    green ;
    BYTE    blue ;
} RgbClr ;


/** ----------------------------------------------------------------------------
    Load MDL application with Name (such as "TSCAN").
    Return 1 if successful.
    Return 0 if application not found.
 */
int LoadApp( char * Name )
{
    void   *Ptr ;
    int     Ok ;

    // Is application already loaded?
    Ptr = mdlSystem_findMdlDesc( Name ) ;
    if ( Ptr )                                            return ( 1 );

    // Not loaded, attempt loading
    Ok = mdlSystem_loadMdlProgram( Name, NULL, "" );
    if ( Ok != SUCCESS ){
        printf( "ERROR: mdlSystem_loadMdlProgram( %s ) = %d - application was not loaded\n", Name, Ok );
        return ( 0 );
    }
    return ( 1 );
}

/** ----------------------------------------------------------------------------
    Call a function in TerraScan that accepts a single parameter of pointer to
    pointer. This function is not tread-safe, since it uses static variables
    for passing by parameter values.
    Set *Ret to be the return value.
    Return 1 if successful.
    Return 0 if could not load TSCAN.
    Return -1 if failed.
 */
int TsCallPptr( int *Ret, const char *FuncName, void **pPtrParam )
{
    CExprValue           Val ;
    CExprResult          Res ;
    int                  Ok ;
    char                 ErrMsg[1024];
    char                 Expr[1024];
    static SymbolSet   * Set = NULL;
    static void        * Ptr = NULL;
    static char        * ParamName = "my_addon_ptrParam";  // should be unique, since is published in the global scope

    if ( !LoadApp( "TSCAN" ) )                            return ( 0 );

    if ( !Set ){
        Set = mdlCExpression_initializeSet( VISIBILITY_CALCULATOR, 0, FALSE );
        if ( !Set ){
            printf( "mdlCExpression_initializeSet() - failed to initialise a set of local variables\n" );
            return ( -1 );
        }
        Ok = mdlCExpression_symbolPublish( Set, ParamName, SYMBOL_CLASS_VAR, mdlCExpression_getType( TYPECODE_POINTER ), & Ptr );
        if ( Ok != SUCCESS ){
            mdlCExpression_generateMessage( ErrMsg, *( toolSubsystem_mdlErrnoP() ) );
            printf( "mdlCExpression_symbolPublish() - failed to publish return variable (status=%d, %s)\n", Ok, ErrMsg );
            return ( -1 );
        }
    }
    Ptr = pPtrParam ? *pPtrParam : NULL;
    sprintf( Expr, "%s( & %s )", FuncName, ParamName );
    Ok = mdlCExpression_getValue( &Val, &Res, Expr, VISIBILITY_CALCULATOR );
    if ( Ok != SUCCESS ){
        mdlCExpression_generateMessage( ErrMsg, *( toolSubsystem_mdlErrnoP() ) );
        printf( "mdlCExpression_getValue( %s ) - failed to publish return variable (status=%d, %s)\n", Expr, Ok, ErrMsg );
        return ( -1 );
    }
    if ( pPtrParam )    *pPtrParam = Ptr;
    if ( Ret )          *Ret = (int) Val.val.valLong ;

    return ( 1 ) ;
}

/** ----------------------------------------------------------------------------
    Incapsulates the call to TerraScan's FnScanGetPnt() function.
 */
int Call_FnScanGetPnt( Point3d **Tbl )
{
    int    Ret;

    return ( TsCallPptr( &Ret, "FnScanGetPnt", Tbl ) > 0 ? Ret : 0 );
}

int Call_FnScanGetCls( BYTE   **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetCls", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetMrk( BYTE   **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetMrk", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetFlg( BYTE   **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetFlg", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetInt( USHORT **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetInt", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetLin( USHORT **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetLin", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetDbl( double **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetDbl", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetEch( BYTE   **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetEch", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetAng( char   **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetAng", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetClr( RgbClr **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetClr", Tbl ) > 0 ? Ret : 0 ); }
int Call_FnScanGetScr( BYTE   **Tbl ){ int Ret; return ( TsCallPptr( &Ret, "FnScanGetScr", Tbl ) > 0 ? Ret : 0 ); }


/** ----------------------------------------------------------------------------
 */
extern int Example( void )
{
    int     Ret;
    Point3d *pPts = NULL;
    BYTE    *pCls = NULL;
    BYTE    *pMrk = NULL;
    BYTE    *pFlg = NULL;
    USHORT  *pInt = NULL;
    USHORT  *pLin = NULL;
    double  *pDbl = NULL;
    BYTE    *pEch = NULL;
    char    *pAng = NULL;
    RgbClr  *pClr = NULL;
    BYTE    *pScr = NULL;

    Ret = Call_FnScanGetPnt( &pPts );  printf( "FnScanGetPnt(): 0x%08X, %d points\n", pPts, Ret );
    Ret = Call_FnScanGetCls( &pCls );  printf( "FnScanGetCls(): 0x%08X, %d points\n", pCls, Ret );
    Ret = Call_FnScanGetMrk( &pMrk );  printf( "FnScanGetMrk(): 0x%08X, %d points\n", pMrk, Ret );
    Ret = Call_FnScanGetFlg( &pFlg );  printf( "FnScanGetFlg(): 0x%08X, %d points\n", pFlg, Ret );
    Ret = Call_FnScanGetInt( &pInt );  printf( "FnScanGetInt(): 0x%08X, %d points\n", pInt, Ret );
    Ret = Call_FnScanGetLin( &pLin );  printf( "FnScanGetLin(): 0x%08X, %d points\n", pLin, Ret );
    Ret = Call_FnScanGetDbl( &pDbl );  printf( "FnScanGetDbl(): 0x%08X, %d points\n", pDbl, Ret );
    Ret = Call_FnScanGetEch( &pEch );  printf( "FnScanGetEch(): 0x%08X, %d points\n", pEch, Ret );
    Ret = Call_FnScanGetAng( &pAng );  printf( "FnScanGetAng(): 0x%08X, %d points\n", pAng, Ret );
    Ret = Call_FnScanGetClr( &pClr );  printf( "FnScanGetClr(): 0x%08X, %d points\n", pClr, Ret );
    Ret = Call_FnScanGetScr( &pScr );  printf( "FnScanGetScr(): 0x%08X, %d points\n", pScr, Ret );

    return ( Ret );
}

/** ----------------------------------------------------------------------------
 */
extern int Example_CloseMousePointAdjustmentWindow( void )
{
    MSWindow   *wnd;
    int        res;

    //    
    printf( "\nStart search..\n" );
    for ( wnd = mdlWindow_getFirst(); wnd; wnd = mdlWindow_getNext( wnd ) ){
        char   wndTitle[256] = "\0";

        //     
        res = mdlWindow_titleGet( wndTitle, sizeof( wndTitle ), wnd );
        printf( "wnd:0x%08x, res=%d, title=%s\n", wnd, res, &wndTitle[0] );
        if ( ( res == SUCCESS )
          && ( mdlstrcmpi( wndTitle, "Mouse Point Adjustment" ) == 0 )
        ){
            //      -   
            break;
        }
    }

    if ( wnd ){
        //  ,     
        printf( "FOUND!\n" );
        mdlWindow_close( wnd, 0, TRUE );
    } else {
        printf( "NOT FOUND!\n" );
    }

    return ( wnd ? 1 : 0 );
}
